from . import conf
from . import utils
import os
splitext = os.path.splitext

langs = set(conf.lang_extensions.keys())

srcext2lang={}
hdrext2lang={}
for lang,exts in conf.lang_extensions.items():
    for hext in exts[0]:
        assert hext not in hdrext2lang
        hdrext2lang['.'+hext]=lang
    for sext in exts[1]:
        assert sext not in srcext2lang
        srcext2lang['.'+sext]=lang

assert set(srcext2lang.keys()).isdisjoint(set(hdrext2lang.keys()))

#todo: use a regexp? (see includes.py for an example)
_header_extensions=set(hdrext2lang.keys())
def headers_in_dir(d):
    return set(utils.listfiles(d,filterfnc=lambda f: splitext(f)[1] in _header_extensions,
                               error_on_no_match=False,ignore_logs=True))



import os
from . import db
from . import col
from . import error
from . import dirs
from . import utils
from . import target_base
join=os.path.join

pypkgname = __name__.split('.')[0]
assert pypkgname != '__main__'

class TargetSymlinks(target_base.Target):
    def __init__(self,pkg,linktype,srcdir,destdir,files,chmodx,renamefct=None,runnables=False):
        db.db['pkg2parts'][pkg.name].add('symlink__%s'%linktype)
        self.name='%s_symlink_%s'%(pkg.name,linktype)
        self.linktype=linktype
        self._cachedir = dirs.pkg_cache_dir(pkg,'symlinks')
        self.files=files
        self.srcdir=srcdir
        self.destdir=destdir
        self.renamefct=renamefct
        self.pkgname=pkg.name
        self.chmodx=chmodx
        pklfile = join(self._cachedir,'%s.pkl'%self.linktype)
        mpf=dirs.makefile_blddir(pklfile)
        self.deps=[mpf]
        self.code = ['@if [ ${VERBOSE} -ge 0 ]; then echo "  %sUpdating symlinks %s/%s%s"; fi'%(col.bldcol('symlink'),
                                                                                                pkg.name,linktype,col.bldend),
                     'python3 -m%s.instsl %s ${VERBOSE}'%(pypkgname,mpf)]

        pklcont=[]
        for f in self.files:
            absf=join(self.srcdir,f)
            assert not os.path.isdir(absf)#should not happen since listfiles ignores subdirs
            if self.chmodx is not None and self.chmodx != utils.is_executable(absf):
                error.error('File must %sbe executable: %s'%('' if self.chmodx else 'not ',os.path.realpath(absf)),
                            '\nYou can correct it by running the command:\n',
                            '    chmod %sx %s'%('+' if self.chmodx else '-',os.path.realpath(absf)))
            dirs.makefile_instdir(self.destdir)
            if self.renamefct:
                r=self.renamefct(pkg,f)
                #dest=dirs.makefile_instdir(self.destdir,r)
            else:
                r=f
                #dest=dirs.makefile_instdir(self.destdir)
            if runnables:
                pkg.register_runnable(r)
                db.db['pkg2runnables'].setdefault(pkg.name,set()).add(r)

            pklcont+=[(f,r)]

        utils.mkdir_p(self._cachedir)
        pklcont = (pklcont,self.srcdir,self.destdir)
        #only update pickle file when contents changed (to avoid triggering makefile targets):
        utils.update_pkl_if_changed(pklcont,pklfile)

def create_tfactory_symlink(filepattern,destdir,chmodx,renamefct=None):
    import re
    match=re.compile(filepattern).match
    destdirabs=str(dirs.installdir / destdir) #converting to string to enable value insertion
    destdir_needs_pkgname = '%s' in destdir
    lf=utils.listfiles
    def tfactory_symlink(pkg,subdir):
        dd=destdirabs%pkg.name if destdir_needs_pkgname else destdirabs
        srcdir=dirs.pkg_dir(pkg,subdir)
        ignore_logs=False
        if subdir=='scripts' or subdir.startswith('app_'):
            ignore_logs=True#handled in the reflogs target
        files=[f for f in lf(srcdir,match,ignore_logs=ignore_logs)]
        if not files:
            error.error("No suitable files found in %s/%s (remove this directory if not needed)"%(pkg.name,subdir))

        return [TargetSymlinks(pkg,subdir,srcdir,dd,files,chmodx=chmodx,renamefct=renamefct,runnables=(subdir=='scripts'))]
    return tfactory_symlink



import os
import errno
import sys
import pathlib

def err(msg):
    #fixme: remove this functions from utils
    from .error import error as errfct
    errfct( msg )

def system(cmd,*,env=None):
    #flush output, to avoid confusing ordering in log-files:
    sys.stdout.flush()
    sys.stderr.flush()
    #rather than os.system, we call "bash -c <cmd>" explicitly through
    #the subprocess module, making sure we can always use bash syntax:
    import subprocess
    cmd=['bash','-c',cmd]
    try:
        ec=subprocess.call(cmd,env=env)
    except KeyboardInterrupt:
        sys.stdout.flush()
        sys.stderr.flush()
        import time
        time.sleep(0.2)
        from . import io as _io
        _io.print("<<<Command execution interrupted by user!>>>")
        if hasattr(sys,'exc_clear'):
            sys.exc_clear()#python2 only
        ec=126
    #wrap exit code to 0..127, in case the return code is passed on to sys.exit(ec).
    sys.stdout.flush()
    sys.stderr.flush()
    if ec>=0 and ec<=127:
        return ec
    return 127


def run(cmd):
    sys.stdout.flush()
    sys.stderr.flush()
    if not isinstance(cmd,list):
        cmd=cmd.split()
    import subprocess
    output=None
    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output = p.communicate()[0]
        ec = p.returncode
    except (OSError,subprocess.SubprocessError):
        ec = 1
    return (ec,output if ec==0 else '')

def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def rm_f(path):
    try:
        os.remove(path)#only succeeds if path exists and is not a directory
    except OSError as e:
        if e.errno == errno.ENOENT:#no such file or directory
            return#not an error, rm -f should exit silently here
        raise

def rm_rf(path):
    #First attempt with os.remove, and only in case it was a directory go for
    #shutil.rmtree (since shutil.rmtree prints error messages if fed symlinks to
    #directories):
    import shutil
    try:
        os.remove(path)
    except OSError as e:
        if e.errno == errno.ENOENT:#no such file or directory
            return#not an error, rm -rf should exit silently here
        elif e.errno != errno.EISDIR and e.errno!=errno.EPERM:
            raise
    #is apparently a directory
    try:
        shutil.rmtree(path)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return#Sudden disappearance is still ok.
        raise

def is_empty_dir( path ):
    #any(..) returns False for empty iterables.
    return path.is_dir() and not any( path.iterdir() )

def rmdir(path):
    p = pathlib.Path(path)
    if p.is_dir() and not any(p.iterdir()):
        p.rmdir()

def listfiles(d,filterfnc=0,error_on_no_match=True,ignore_logs=False):
    from . import conf
    for f in os.listdir(d):
        if ignore_logs and f.endswith('.log'):
            continue
        if not conf.ignore_file(f) and not os.path.isdir(os.path.join(d,f)):
            if filterfnc:
                if filterfnc(f):
                    yield f
                elif error_on_no_match:
                    from . import error
                    error.error("Forbidden file %s/%s"%(d,f))
            else:
                yield f

def is_executable(fn):
    return os.access(fn,os.X_OK)

def pkl_load(fn_or_fh):
    import pickle
    if hasattr(fn_or_fh,'read'):
        #argument is filehandle:
        return pickle.load(fn_or_fh)#filehandle already
    else:
        #argument is filename:
        with pathlib.Path(fn_or_fh).open('rb') as fh:
            return pickle.load(fh)

def pkl_dump(data,fn_or_fh):
    import pickle
    if hasattr(fn_or_fh,'write'):
        #argument is filehandle:
        d=pickle.dump(data,fn_or_fh)
        fn_or_fh.flush()
        return d
    else:
        #argument is filename:
        with pathlib.Path(fn_or_fh).open('wb') as fh:
            d=pickle.dump(data,fh)
        return d

def update_pkl_if_changed(pklcont,filename):
    old = os.path.exists(filename)
    readold = False
    oldcont = None
    changed = True
    if old:
        try:
            oldcont=pkl_load(filename)
            readold = True
        except EOFError:
            from . import io as _io
            _io.print( "WARNING: Old pickle file %s ended unexpectedly"%filename)
            oldcont = None
            readold = False
        if readold and oldcont == pklcont:
            changed = False
    if not changed:
        return oldcont
    if old:
        os.rename(filename,str(filename)+'.old')
    pkl_dump(pklcont,filename)
    return oldcont

def shlex_split(s):
    try:
        import shlex
    except ImportError:
        shlex=None
    if shlex:
        return shlex.split(s)
    else:
        return s.split()#hope user didnt use spaces in directories

class rpath_appender:
    def __init__( self, lang, shlib ):
        from . import env
        langinfo=env.env['system']['langs'][lang]
        rpath_pattern = langinfo['rpath_flag_lib' if shlib else 'rpath_flag_exe']
        self.__patterns = [ rpath_pattern ]
        if langinfo['can_use_rpathlink_flag'] and '-rpath' in rpath_pattern and '-rpath-link' not in rpath_pattern:
            self.__patterns += [ self.__patterns[0].replace('-rpath','-rpath-link') ]
    def apply_to_dir( self, directory ):
        return [ p%directory for p in self.__patterns ]
    def apply_to_flags( self, flag_list ):
        found_dirs = []
        for f in flag_list:
            for p in ['-L','-Wl,rpath,','-Wl,rpath=','-Wl,rpath-link,','-Wl,rpath-link=']:
                _ = None
                if f.startswith(p):
                    _ = f[len(p):]
                    if not _:
                        continue
                elif not f.startswith('-'):
                    _ = pathlib.Path(f)
                    if _.exists() and not _.is_dir() and _.parent.is_dir():
                        _ = str(_.parent.absolute().resolve())#../bla/libfoo.so -> /some/where/bla
                    elif not _.exists():
                        _ = None
                _ = str(_) if _ else None
                if _ and _ not in found_dirs:
                    found_dirs.append( _ )
        res = flag_list[:]
        for d in found_dirs:
            res += self.apply_to_dir( d )
        return res

def path_is_relative_to( p, pother ):
    #Path.is_relative_to(..) was introduced in Python 3.9, this function lets us
    #support python 3.8.
    assert isinstance( p, pathlib.Path )
    if hasattr(p,'is_relative_to'):
        #Python 3.9+:
        return p.is_relative_to(pother)
    else:
        #Python 3.8:
        try:
            p.relative_to(pother)
            return True
        except ValueError:
            return False

def _remove_symlinks( symlinks, verbose ):
    #FIXME: pathlib
    dirs_with_removals=set()#to test for empty dirs to be removed at the end
    if verbose:
        from .io import print
    else:
        def print( *a, **kw ):
            return None
    for fn in symlinks:
        print("Removing obsolete symlink %s"%fn)
        try:
            os.remove(fn)
        except OSError as exc:
            if exc.errno == errno.ENOENT and not os.path.exists(fn):
                #perhaps we were aborted and restarted so we should just report and carry on
                print("WARNING: Obsolete symlink already removed: %s"%fn)
                pass
            else:
                raise
        dirs_with_removals.add( os.path.dirname( fn ) )

    #check for and cleanup empty dirs:
    if dirs_with_removals:
        #we removed files, remove dirs if not empty:
        def isemptydir( p ):
            return not any( pathlib.Path(p).iterdir() )
        for d in (d for d in dirs_with_removals if isemptydir(d)):
            try:
                os.rmdir(d)
            except OSError as exc:
                if exc.errno == errno.ENOENT:
                    pass#could have been already been removed by other target due to race condition
                else:
                    raise


def update_symlinks( oldlist, newlist, verbose ):
    if verbose:
        from .io import print
    else:
        def print( *a, **kw ):
            return None

    obsoleted = oldlist.difference(newlist)
    _remove_symlinks( [ os.path.join(destdir,linkname)
                        for destdir,linkname,src in obsoleted ],
                      verbose )
    #create dirs:
    for dd in set(d for d,_,_ in newlist):
        mkdir_p(dd)
    #create links:
    for destdir,linkname,src in newlist:
        fn = os.path.join( destdir, linkname )
        if os.path.lexists(fn):
            if os.path.samefile( fn, src ):
                #Already in place
                continue
            else:
                try:
                    os.unlink(fn)
                except FileNotFoundError:
                    #Removed in race condition?
                    pass
        print("Installing symlink %s"%fn)
        try:
            os.symlink(src,fn)
        except OSError as exc:
            if exc.errno == errno.EEXIST:
                pass#could have been created previously before an abort
            else:
                raise



def main():
    import sys
    import errno
    from . import utils
    fn=sys.argv[1]
    verbose=int(sys.argv[2]) if len(sys.argv)>=3 else 0
    needed=utils.pkl_load(fn)
    ofn='%s.old'%fn
    try:
        current = utils.pkl_load(ofn)
    except EOFError:
        current = set()
    except IOError as e:
        if e.errno == errno.ENOENT:#no such file or directory
            current = set()
        else:
            raise
    utils.update_symlinks( current, needed, verbose )

if __name__=='__main__':
    import sys
    try:
        main()
    except KeyboardInterrupt:
        from .io import print
        print("<<symlink installation interrupted by user!>>>")
        #Fixme: Any way to recover state?? (if so we should also catch other errors)
        sys.stdout.flush()
        sys.stderr.flush()
        import time
        time.sleep(0.2)
        sys.exit(1)



def find( pkg, pattern, filenames = None ):
    import os
    import fnmatch
    import pathlib
    from .grep import pkgfiles
    from . import col
    pkgcol = col.ok if pkg.enabled else col.bad

    n=0
    pkg_rd = pkg.reldirname#always match on package reldir
    pattern='*%s*'%pattern.lower()
    from .io import print
    cwd = pathlib.Path.cwd()
    for flocal in pkgfiles(pkg):
        flocal = pathlib.Path(flocal)
        if filenames and flocal not in filenames:
            continue
        if fnmatch.fnmatch( str(pkg_rd / flocal).lower(), pattern ):
            #os.path.relpath does not have pathlib equivalent here
            print( '   %s%s%s'%( pkgcol,
                                 os.path.relpath((pkg.dirname / flocal),cwd),
                                 col.end ) )
            n+=1

    return n




# Various utility functions to be used in the sphinx doc conf.py in both
# simplebuild itself, but also related projects.

def _get_gitversion( reporoot ):
    import subprocess
    #--always to not fail in a shallow clone:
    cmd=[ 'git','describe','--tags','--always',
          '--match','v[0-9]*.[0-9]*.[0-9]*' ]
    import os
    if os.environ.get('SIMPLEBUILD_GITVERSION_USE_LATEST_VTAG'):
        cmd += [ '--abbrev=0' ]
    p = subprocess.run( cmd,
                        cwd = reporoot,
                        capture_output = True )
    if p.returncode != 0:
        print(p.stderr.decode())
        print(p.stdout.decode())
        raise RuntimeError( f'Command "{" ".join(cmd)}" '
                            f'in dir {reporoot} failed!' )

    assert not p.stderr
    return p.stdout.decode()

def guess_language( path ):
    if path.name=='pkg.info':
        return 'pkginfo'
    if path.name=='simplebuild.cfg':
        return 'toml'
    suf = path.suffix
    if suf in ('.cc','hh','icc'):
        return 'c++'
    if suf in ('.f',):
        return 'fortran'
    if suf in ('.c','h'):
        return 'c'
    if suf in ('.py',):
        return 'python'
    if not suf:
        lines = path.read_text().splitlines()
        if lines and lines[0].startswith('#!/'):
            return {
                '#!/usr/bin/env python3' : 'python',
                '#!/usr/bin/env bash' : 'bash',
            }.get(lines[0])

def get_display_language(language):
    if language == 'pkginfo':
        return ''
    if language in ('toml','bash'):
        return language.upper()
    return language.capitalize()

_fixuptest_dirrepl = []
def add_fixuptext_dirreplacement( adir, text ):
    _fixuptest_dirrepl.append( ( adir, text ) )

def fixuptext( pkgroot, txt ):
    import _simple_build_system
    import pathlib
    import os
    sbsdir = pathlib.Path(_simple_build_system.__file__).parent
    dirrepl = (_fixuptest_dirrepl or [])[:]
    dirrepl.append( ( pkgroot.parent, '/some/where/' ) )
    dirrepl.append( ( sbsdir.parent, '/some/where/else/' ) )

    cp = os.environ.get('CONDA_PREFIX')
    if cp:
        dirrepl.append( ( pathlib.Path(cp), '/conda/envs/sbenv/' ) )

    for p, t in dirrepl:
        txt = txt.replace( str(p.absolute())+'/', t )

    return txt

def check_output_contains( textfile, pattern, must_contain = True ):
    if (pattern in textfile.read_text()) == must_contain:
        return
    issue = 'does not contain' if must_contain else 'contains forbidden'
    print(f'\n\n\nERROR: File {textfile} {issue} pattern "{pattern}":\n')
    for e in textfile.read_text().splitlines(keepends=True):
        print(f'>>>{e}',end='')
    raise SystemExit(1)

def check_output_not_contains( textfile, pattern ):
    check_output_contains( textfile, pattern, must_contain = False )




#configuration file (warning: not everything can be changed here without updating other modules as well)

import os
import pathlib
import sysconfig
from . import io as _io

def AbsPath( p ):
    return pathlib.Path(p).expanduser().resolve().absolute()

__cache_pyextsuf = sysconfig.get_config_var('EXT_SUFFIX') #value like '.cpython-311-x86_64-linux-gnu.so' (.so even on osx)

lang_extensions = {
    'cxx' : ( ['hh', 'icc'], ['cc'] ),
    'c' : ( ['h'], ['c'] ),
    'fortran' : ( [], ['f'] ),
}

package_cfg_file='pkg.info'

autodeps = set(['Core'])
projectname='SBLD'
projectname_lc = projectname.lower()
runnable_prefix = 'sb_'

#TODO: Do not repeat .io stuff here:
print_prefix_name = _io.print_prefix_name
print_prefix = _io.print_prefix
def print( *args, **kwargs ):
    return _io.print( *args, **kwargs )
def print_no_prefix( *args, **kwargs ):
    return _io.print_no_prefix( *args, **kwargs )
def make_quiet():
    return _io.make_quiet()

def runnable_name(pkg,base_name):
    #create global runnable name for a runnable in a package:
    return ('%s%s_%s'%(runnable_prefix,pkg.name,base_name)).lower()

def runnable_is_test(runnable_name):
    return runnable_name.split('_',2)[2].startswith('test')

def libldflag(pkg): return '-lPKG__%s'%pkg.name
def namefct_lib(pkg,subdir,platform_pattern): return platform_pattern%('PKG__'+pkg.name)
def namefct_pycpp(pkg,subdir,platform_pattern): return subdir[6:] + __cache_pyextsuf

def namefct_app(pkg,subdir,platform_pattern): return runnable_name(pkg,subdir[4:])

def checkfct_pycpp(pkg,subdir,name):
    n=subdir[6:]
    if n=='__init__':
        return 'The name pycpp___init__/ is forbidden. Instead put compiled init code in a subdir called pycpp__init/'
    cf=os.path.join(pkg.dirname,'python/%s.py'%n)
    if os.path.exists(cf):
        return 'Python module %s.%s is provided in both pure (python/%s.py) and compiled (pycpp_%s/) forms'%(pkg.name,n,n,n)

def descrfct_lib(pkg,subdir,name):
    from . import col
    return col.bldcol('shlib'),'shared library for package %s'%pkg.name

def descrfct_pycpp(pkg,subdir,name):
    from . import col
    return col.bldcol('pymod'),'python module %s.%s'%(pkg.name,subdir[6:])

def descrfct_app(pkg,subdir,name):
    from . import col
    return col.bldcol('app'),'application %s'%name

def uninstall_package(pkgname):
    #completely remove all traces of a pkg from the install area:
    #a few sanity checks since we are about to use rm -rf:
    #assert d and not ' ' in d

    instdir = install_dir()
    if not ( instdir / '.sbinstalldir' ).exists():
        return

    #FIXME: <pn>_foo_bar_blah might be script Blah from package Foo_Bar or script
    #Bar_Blah from package Foo. We should check that the symlinks goes to the
    #correct package! (or better yet, simplebuild should produce pickle file in
    #install or bld with all package dependencies and provided scripts, apps,
    #etc.). But a quick fix for the scripts (not for the apps) would be for the
    #framework to remove all symlinks to the package in question from the
    #install dir, and just let the present function deal with non-symlinks.

    parts = [ f'data/{pkgname}',
              f'lib/*PKG__{pkgname}.*',
              f'tests/testref/{runnable_prefix}{pkgname.lower()}_*.log',
              f'include/{pkgname}',
              f'python/{pkgname}',
              f'scripts/{runnable_prefix}{pkgname.lower()}_*',
              f'bin/{runnable_prefix}{pkgname.lower()}_*' ]
    import shutil
    for p in parts:
        for f in instdir.glob(p):
            if f.is_dir():
                shutil.rmtree(f,ignore_errors = True)
            else:
                f.unlink(missing_ok = True)

#Get paths to all packages (including the framework and user packages)

def projects_dir():
    from . import envcfg
    return envcfg.var.projects_dir

def extra_pkg_path():
    from . import envcfg
    return envcfg.var.extra_pkg_path_list

def pkg_search_path():
    #candidates = [framework_dir(), projects_dir()]
    candidates = [projects_dir()]
    candidates.extend(extra_pkg_path())
    dirs = []
    for d in candidates:
        if d not in dirs:
            dirs.append( d )
    return dirs

def build_dir():
    from . import envcfg
    return envcfg.var.build_dir_resolved

def install_dir():
    from . import envcfg
    return envcfg.var.install_dir_resolved

def test_dir():
    return build_dir() / 'testresults/'

def safe_remove_install_and_build_dir():
    import shutil
    for fingerprintfile in [ build_dir() / '.sbbuilddir', install_dir() / '.sbinstalldir' ]:
        if fingerprintfile.exists():
            shutil.rmtree( fingerprintfile.parent, ignore_errors=True )

def target_factories_for_patterns():
    from . import tfact_symlink as tfs
    from . import tfact_headerdeps as tfh
    from . import tfact_binary as tfb
    l=[]
    l += [('data',   tfs.create_tfactory_symlink('','data/%s',chmodx=False))]
    l += [('scripts',tfs.create_tfactory_symlink('','scripts',chmodx=True,
                                                 renamefct=runnable_name))]#todo: disallow periods? enforce lowercase?
    l += [('python', tfs.create_tfactory_symlink('.+\.py','python/%s',chmodx=False))]
    l += [('pycpp_.+', tfb.create_tfactory_binary(shlib=True,
                                                  instsubdir='python/%s',
                                                  allowed_langs=['cxx'],
                                                  namefct=namefct_pycpp,
                                                  descrfct=descrfct_pycpp,
                                                  checkfct=checkfct_pycpp,
                                                  flagfct=lambda pkg,subdir:['-DPYMODNAME=%s'%subdir[6:]]))]
    l += [('app_.+', tfb.create_tfactory_binary(instsubdir='bin',namefct=namefct_app,
                                                  descrfct=descrfct_app))]
    l += [('libsrc', tfb.create_tfactory_binary(pkglib=True,namefct=namefct_lib,descrfct=descrfct_lib))]
    l += [('libinc',  tfh.tfactory_headerdeps)]#just for header dependencies


    return l

def ignore_file(f):
    f = f.name if hasattr(f,'name') else f
    return f[0]=='.' or '~' in f or '#' in f or f.endswith('.orig') or f.endswith('.bak') or f=='__pycache__'

def target_factories():
    #non-pattern factories:
    from . import tfact_prepinc
    from . import tfact_pyinit
    from . import tfact_reflogs
    from . import tfact_libavail
    l = []
    l += [tfact_prepinc.tfactory_prepinc]
    l += [tfact_pyinit.tfactory_pyinit]
    l += [tfact_reflogs.tfactory_reflogs]
    l += [tfact_libavail.tfactory_libavail]
    return l

def deinstall_parts(instdir,pkgname,current_parts,disappeared_parts):
    from . import dirs
    i=instdir
    if not ( i / '.sbinstalldir' ).exists():
        return
    unused=set()
    pydone=False
    pkgcache=dirs.pkg_cache_dir(pkgname)
    import shutil
    def rm_tree(p):
        shutil.rmtree( p, ignore_errors=True)
    def rm_file( p ):
        p.unlink(missing_ok = True)
    def rm_pattern(thedir,pattern):
        for f in thedir.glob(pattern):
            rm_file(f)

    for d in disappeared_parts:
        if d=='libinc':
            rm_tree( i / 'include' / pkgname)
        elif d=='libsrc':
            rm_pattern(i/'lib','*PKG__%s.*'%pkgname)
        elif d.startswith('app_'):
            _runnable_name='%s%s_%s' %( runnable_prefix,
                                        pkgname.lower(),
                                        d[4:].lower() )
            rm_tree( i / 'bin' / _runnable_name )
        elif d=='symlink__scripts':
            #FIXME: Next line clashes (see FIXME above)
            rm_pattern( i/'scripts',
                        '%s%s_*'%(runnable_prefix,pkgname.lower()) )
            (pkgcache/'symlinks'/'scripts.pkl').touch()
            rm_file(pkgcache/'symlinks'/'scripts.pkl.old')
        elif d=='symlink__data':
            rm_tree( i / 'data' / pkgname )
            ( pkgcache / 'symlinks' / 'data.pkl' ).touch()
            rm_file( pkgcache / 'symlinks' / 'data.pkl.old' )
        #don't do this for testref_links, since all packages always have this target:
        #elif d=='testref_links':
        #    NB: Syntax not updated for pathlib and fcts above!:
        #    utils.rm_f(os.path.join(i,'tests/testref/%s%s_*.log'%(runnable_prefix,pkgname.lower())))
        #    utils.touch(os.path.join(pkgcache,'testref/testref.pkl'))
        #    utils.rm_f(os.path.join(pkgcache,'testref/testref.pkl.old'))
        elif d.startswith('autopyinit'):
            if not any(e.startswith('autopyinit') for e in current_parts):
                #we must remove the auto generated __init__.py as it is in the way
                rm_file( i / 'python' / pkgname / '__init__.py' )
                ( pkgcache / 'symlinks' / 'python.pkl' ).touch()
                rm_file( pkgcache / 'symlinks' / 'python.pkl.old' )
        elif d.startswith('pycpp_') or d=='symlink__python':
            if d.startswith('pycpp'):
                rm_file( ( i / 'python' / pkgname / '%s.so' )%d[6:] )
            if not pydone:
                pydone=True
                if not any((e.startswith('pycpp') or e=='symlink__python') for e in current_parts):
                    rm_tree( i / 'python' / pkgname )
                ( pkgcache / 'symlinks' / 'python.pkl' ).touch()
                rm_file( pkgcache / 'symlinks' / 'python.pkl.old' )
                ( pkgcache / 'pyinit' / 'pyinit.pkl' ).touch()
                #rm_file( pkgcache / 'pyinit' / 'pyinit.pkl.old' )
        else:
            unused.add(d)
    assert not unused,str(unused)



import os
import pathlib
from . import utils
from . import langs
from . import col
from . import includes
from . import conf
from .grep import pkgdir_for_search
_ospath = os.path

ccxx_extensions = set(ext for ext,lang in langs.hdrext2lang.items() if lang in ('cxx','c'))
ccxx_extensions.update(set(ext for ext,lang in langs.srcext2lang.items() if lang in ('cxx','c')))

def ccxx_files_in_dir(d):
    out=[]
    for f in os.listdir(d):
        if not conf.ignore_file(f):
            if _ospath.splitext(f)[1] not in ccxx_extensions:
                continue
            f=_ospath.join(d,f)
            if _ospath.isdir(f):
                continue
            out+=[f]
    return out

def _printform(pkg,f,for_sortkey=False):
    f = pathlib.Path(f).parts[-3:]
    assert f[0]==pkg.name
    if for_sortkey:
        return tuple(f)
    return '%s%s/%s/%s%s'%(col.ok if pkg.enabled else col.bad,
                           f[0],f[1],f[2],col.end)

def _rel_description(f,f2):
    f=_ospath.dirname(f)
    f2=_ospath.dirname(f2)
    if f==f2:
        return ' %s(same directory)%s'%(col.inc_samedir,col.end)
    f=_ospath.dirname(f)
    f2=_ospath.dirname(f2)
    if f==f2:
        return ' %s(same package)%s'%(col.inc_samepkg,col.end)
    return ''

_cache_incs_direct={}
def _incs_direct(pkg,f):
    global _cache_incs_direct
    if f in _cache_incs_direct:
        return _cache_incs_direct[f]
    res={}
    possible_privincs,possible_pkgincs = includes.find_includes(f,pkg)
    d=_ospath.dirname(f)
    if possible_privincs:
        for p in possible_privincs:
            pp=_ospath.join(d,p)
            if _ospath.exists(pp) and pp!=f:
                res[_printform(pkg,pp,for_sortkey=True)] = (pkg,pp)
    if possible_pkgincs:
        deppkgs = dict((p.name,p) for p in pkg.deps())
        deppkgs[pkg.name]=pkg#also depend on ourselves here
        for ppkgname,p in possible_pkgincs:
            ppkg=deppkgs.get(ppkgname,None)
            if not ppkg:
                continue
            assert ppkg.name==ppkgname
            p=pkgdir_for_search(ppkg,'libinc',p)
            if _ospath.exists(p) and _ospath.realpath(p)!=_ospath.realpath(f):
                res[_printform(ppkg,p,for_sortkey=True)] = (ppkg,p)
    _cache_incs_direct[f]=res
    return res

def _incs_recursive(pkg,f,out):
    for k,v in _incs_direct(pkg,f).items():
        if k in out:
            continue
        out[k]=v
        _incs_recursive(v[0],v[1],out)
    return out

_cache_incs_indirect={}
def _incs_indirect(pkg,f):
    global _cache_incs_indirect
    if f in _cache_incs_indirect:
        return _cache_incs_indirect[f]
    out={}
    _incs_recursive(pkg,f,out)
    for k,v in _incs_direct(pkg,f).items():
        del out[k]
    _cache_incs_indirect[f]=out
    return out

def collect_info(pkgloader,f):
    #collects info from single file and returns in dict.
    pkgname,subdir,fn = f.split(_ospath.sep)[-3:]
    pkg = pkgloader.name2pkg.get(pkgname,None)
    if not pkg or not f.startswith(pkg.dirname):
        utils.err('File %s not located inside a package'%f)
    #find isheader,lang from extension:
    _,ext = _ospath.splitext(f)
    isheader=True
    lang = langs.hdrext2lang.get(ext,None)
    if lang is None:
        isheader=False
        langs.srcext2lang.get(ext,None)

    lang = langs.srcext2lang.get(ext,langs.hdrext2lang.get(ext,None))
    if lang not in ['cxx','c']:
        utils.err('File %s not a C++ or C file'%f)

    nfo = dict(signature=(pkgname,subdir,fn),
               pkg=pkg,
               lang=lang,
               isheader=isheader,
               includes_indirectly=[],is_included_by_indirectly=[],
               includes_directly=[],is_included_by_directly=[])
    for _,(pi,fi) in sorted([(k,v) for k,v in _incs_direct(pkg,f).items()]):
        nfo['includes_directly'] += [(pi,fi)]
    for _,(pi,fi) in sorted([(k,v) for k,v in _incs_indirect(pkg,f).items()]):
        nfo['includes_indirectly'] += [(pi,fi)]
    #Now the harder part, include statements TO the file in question:
    if not isheader:
        return nfo
    #This is a header file, so in other words in can be included by other
    #files. First get the list of files which could possibly include the file in question.
    if subdir!='libinc':
        #only consider files in the same directory
        files = [(pkg,ff) for ff in ccxx_files_in_dir(pkgdir_for_search(pkg,subdir))]
    else:
        #the hardest task: consider files in all dependent packages
        files=[]
        pkgs=set(pkg.all_clients())
        pkgs.add(pkg)
        for pp in pkgs:
            for d in os.listdir(pkgdir_for_search(pp)):
                d=pkgdir_for_search(pp,d)
                if _ospath.isdir(d):
                    files += [(pp,ff) for ff in ccxx_files_in_dir(d)]
    files.remove((pkg,f))#not ourself
    files = [(_printform(pp,ff,for_sortkey=True),pp,ff) for pp,ff in files]
    files.sort()

    #ok, now we have a list of files which COULD include us, let us see who does:
    searchkey = _printform(pkg,f,for_sortkey=True)
    for _,pp,ff in files:
        if searchkey in _incs_direct(pp,ff):
            nfo['is_included_by_directly'] += [(pp,ff)]
    for _,pp,ff in files:
        if searchkey in _incs_indirect(pp,ff):
            nfo['is_included_by_indirectly'] += [(pp,ff)]
    return nfo

def _classify(nfo):
    return '%s %s'%({'cxx':'C++','c':'C'}[nfo['lang']],
                    "header" if nfo['isheader'] else "source")

def provide_info(pkgloader,f):
    from .io import print
    nfo = collect_info(pkgloader,f)
    print()
    print("Investigating %s file: %s"%(_classify(nfo),_printform(nfo['pkg'],f)))
    nrelationships = 0
    for i in ('includes','is included by'):
        for d in ('directly','indirectly'):
            l=['%s%s'%(_printform(pi,fi),_rel_description(f,fi)) for (pi,fi) in sorted(nfo['%s_%s'%(i.replace(' ','_'),d)])]
            if l:
                nrelationships += len(l)
                t='1 file' if len(l)==1 else '%i files'%len(l)
                print()
                print("File %s %s (%s):"%(i,t,d))
                print()
                for e in l:
                    print('      %s'%e)
    if not nrelationships:
        print()
        print("File neither includes nor is included by any other file in the framework")
    print()

def provide_info_multifiles(pkgloader,files):
    from .io import print
    nfos = dict((fn,collect_info(pkgloader,fn)) for fn in files)
    print()
    print("Investigating %i files:"%len(files))
    print()
    for fn in files:
        nfo = nfos[fn]
        print('      %s (%s)'%(_printform(nfo['pkg'],fn),_classify(nfo)))
    #find merged includes:
    merged_nfo = {}
    files_set = set(files)#ignore dependencies within the queried list of files
    for key in ('includes_directly', 'includes_indirectly', 'is_included_by_directly', 'is_included_by_indirectly'):
        seen,l=set(),[]
        for nfo in nfos.values():
            l += [v for v in nfo[key] if v[1] not in files_set and not (v in seen or seen.add(v))]#relies on seen.add returning None
        merged_nfo[key] = l
    #print results
    nrelationships = 0
    for i in ('includes','is included by'):
        for d in ('directly','indirectly'):
            l=[_printform(pi,fi) for (pi,fi) in sorted(merged_nfo['%s_%s'%(i.replace(' ','_'),d)])]
            if l:
                i_plural = {'includes':'include','is included by':'are included by'}[i]
                nrelationships += len(l)
                t='1 file' if len(l)==1 else '%i files'%len(l)
                print()
                print("Files %s %s (%s):"%(i_plural,t,d))
                print()
                for e in l:
                    print('      %s'%e)
    if not nrelationships:
        print()
        print("Files neither includes nor are included by other files in the framework")
    print()




_cache_version = [None]
def determine_version():
    if _cache_version[0] is None:
        import importlib.metadata
        try:
            #NB: Hardwiring python package name:
            version = importlib.metadata.version('simple-build-system')
        except importlib.metadata.PackageNotFoundError:
            # package is not installed (not really supported)
            version = 'unversioned_local_source'
        _cache_version[0] = version
    return _cache_version[0]



def formatlist(l,width,colour=None,indent_first='',indent_others='',nonetext='<none>'):
    from . import col
    #entries in input list are either in format (text,color) or simply
    #(text). colour will be used to override all colours (colour='' will turn
    #all colours off).
    out=['']
    i=0
    for entry in l:
        try:
            e,c = entry
        except ValueError:
            e=entry
            c=colour
        if colour is not None:
            c=colour
        if i>width:
            out+=['']
            i=0
        k=len(e)
        if c:
            e=c+e+col.end
        out[-1]+=' '+e if i else e
        i+=k
    if nonetext and out==['']:
        out=[nonetext]
    for i in range(len(out)):
        if indent_first and i==0:
            out[i]=indent_first+out[i]
        if indent_others and i>0:
            out[i]=indent_others+out[i]
    return out



from . import dirs
from . import target_base
import os.path
from . import utils
from . import col
from . import db
join=os.path.join

pypkgname = __name__.split('.')[0]
assert pypkgname != '__main__'

class TargetPrepInc(target_base.Target):

    def __init__(self,pkg,has_compiled_target,has_libinc):
        self.pkgname=pkg.name
        self.name='%s_prepinc'%pkg.name
        self._privincdir=dirs.pkg_cache_dir(pkg,'inc')
        self._has_libinc=has_libinc
        self._has_compiled_target=has_compiled_target
        self._cachedir=dirs.pkg_cache_dir(pkg,'prepinc')
        self._pklfile = join(self._cachedir,'prepinc.pkl')
        mpf=dirs.makefile_blddir(self._pklfile)
        self.deps=[mpf]
        self.code = ['python3 -m%s.instsl2 %s ${VERBOSE}'%(pypkgname,mpf)]
        utils.mkdir_p(self._cachedir)
        self._neededlinks=set()
        if self._has_libinc:
            #contains_message=True
            self.code.insert(0,'@if [ ${VERBOSE} -ge 0 ]; then echo "  %sInstalling %s headers%s"; fi'%(col.bldcol('headers'),pkg.name,col.bldend))
            db.db['pkg2parts'][pkg.name].add('libinc')
            li=dirs.pkg_dir(pkg,'libinc')
            self._neededlinks.add((dirs.installdir / dirs.incdirname,pkg.name,li))
            if self._has_compiled_target:
                self._neededlinks.add((self._privincdir,pkg.name,li))
        #we also use this target to keep the DB of available public headers up to date (needed for header deps)
        pkg.update_incs()

    def setup(self,pkg):
        if self._has_compiled_target:
            for p in pkg.deps():
                d=dirs.pkg_dir(p,'libinc')
                if os.path.isdir(d):
                    self._neededlinks.add((self._privincdir,p.name,d))
        #only update pickle file when contents changed (to avoid triggering makefile targets):
        utils.update_pkl_if_changed(self._neededlinks,self._pklfile)

_compiled_patterns = set(['pycp', 'app_', 'libs'])#todo: to conf
def tfactory_prepinc(pkg,dirtypes):
    has_compiled_target = not dirtypes.isdisjoint(_compiled_patterns)
    has_libinc = 'libi' in dirtypes#the right way, but for now we use the following:
    if has_compiled_target or has_libinc:
        return [TargetPrepInc(pkg,has_compiled_target,has_libinc)]
    else:
        return []



#global variables
import pathlib
import os
from . import conf#project specific configuration

#system dir is one up from the modules dir:
sysdir = str(pathlib.Path(__file__).resolve().absolute().parent.parent)
blddir = conf.build_dir() # imports envcfg
makefiledir = blddir / 'makefiles'

extrapkgpath = conf.extra_pkg_path() # imports envcfg
pkgsearchpath = conf.pkg_search_path() # imports envcfg

installdir = conf.install_dir() # imports envcfg
testdir = conf.test_dir() # imports envcfg
projdir = conf.projects_dir() # imports envcfg
datadir = pathlib.Path(__file__).resolve().absolute().parent / 'data'
cmakedetectdir = datadir / 'cmake'

incdirname='include'
#libdirname='lib'#fixme: unused, 'lib' is simply hardcoded in a few places.

envcache = blddir / 'env.cache'
systimestamp_cache=blddir / 'systimestamp.cache'
lockfile=blddir / ".lock"

def makefile_instdir(*subpaths):
    if not subpaths:
        return "${INST}"
    subpaths=os.path.join(*subpaths)
    if subpaths[0]=='/':
        subpaths=os.path.relpath(subpaths,installdir)
    return os.path.join("${INST}",subpaths)

def makefile_blddir(*subpaths):
    if not subpaths:
        return "${BLD}"
    subpaths=os.path.join(*subpaths)
    if subpaths[0]=='/':
        subpaths=os.path.relpath(subpaths,blddir)
    return os.path.join("${BLD}",subpaths)

def _pkgname(pkg):
    return pkg.name if hasattr(pkg,'name') else pkg

def pkg_cache_dir(pkg,*subpaths):
    return blddir.joinpath('pc',_pkgname(pkg),*subpaths)

def makefile_pkg_cache_dir(pkg,*subpaths):
    return os.path.join('${BLD}','pc',_pkgname(pkg),*subpaths)

#where we link (or create dynamic pkgs):
pkgdirbase = blddir / 'pkgs'
def pkg_dir(pkg,*subpaths):
    return pkgdirbase.joinpath( _pkgname(pkg),*subpaths )

def makefile_pkg_dir(pkg,*subpaths):
    return os.path.join('${PKG}',_pkgname(pkg),*subpaths)

#sanity:
for d in [str(x) for x in [blddir, *pkgsearchpath, installdir]]:
    assert ' ' not in d, 'Spaces not allowed in directory names. Illegal path is: "%s"'%d
    assert len(d)>3,f"suspiciously short path name: {d}"

# Package directory aliases #keep them lowercase
pkgdir_aliases = {
  "projects": projdir,
  "extra": extrapkgpath
  }

def create_bld_dir():
    fingerprint = blddir  / '.sbbuilddir'
    if  blddir.exists() and not fingerprint.exists():
        from . import error
        error.error(f'Did not find expected fingerprint file at: {fingerprint}')
    blddir.mkdir(parents=True,exist_ok=True)
    fingerprint.touch()
    assert blddir.is_dir()
    assert ( blddir  / '.sbbuilddir' ).exists()

def create_install_dir():
    fingerprint = installdir  / '.sbinstalldir'
    if  installdir.exists() and not fingerprint.exists():
        from . import error
        error.error(f'Did not find expected fingerprint file at: {fingerprint}')
    installdir.mkdir(parents=True,exist_ok=True)
    fingerprint.touch()
    assert installdir.is_dir()
    assert ( installdir  / '.sbinstalldir' ).exists()



import pathlib
import typing
import re
from . import error

_OptPath = typing.Union[pathlib.Path, None]# "pathlib.Path | None" needs py3.10+

class TOMLSchemaDecodeContext(typing.Protocol):
    @property
    def src_descr(self) -> str:
        return ''
    @property
    def default_dir(self) -> _OptPath:
        return None
    @property
    def item_name(self) -> str:
        return ''

def decode_list_of_search_paths( ctx : TOMLSchemaDecodeContext, item ):
    if not isinstance( item, list ):
        error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected list) in {ctx.src_descr}')
    res=[]
    for i,k in enumerate(item):
        if not isinstance(k,str) or not k:
            error.error(f'Invalid value of list entry #{i+1} in item {ctx.item_name} (expected non-empty string) in {ctx.src_descr}')
        p = pathlib.Path(k).expanduser()
        if not p.is_absolute():
            if ctx.default_dir is None:
                error.error(f'Invalid option "{item}" for item {ctx.item_name} (relative paths can only'
                            f' be used in a context where the "root" dir is apparent) in {ctx.src_descr}')
            p = ctx.default_dir / p
        if not p.exists():
            error.error(f'Non-existing path "{p}" in list entry #{i+1} in item {ctx.item_name} in {ctx.src_descr}')
        if p.is_file():
            p_redirect = p if p.name=='simplebuild_redirect.cfg' else None
            p_std = p if p_redirect is None else None
        if p.is_dir():
            #Unless the filename is specified directly, add the default name:
            p_std = ( p / 'simplebuild.cfg' )
            p_redirect = ( p / 'simplebuild_redirect.cfg' )
            p_std = None if not p_std.exists() else p_std
            p_redirect = None if not p_redirect.exists() else p_redirect
        if p_std is None and p_redirect is None:
            error.error(f'Missing file "{p}" in list entry #{i+1} in item {ctx.item_name} in {ctx.src_descr}')
        if p_std is not None and p_redirect is not None:
            error.error('Both simplebuild.cfg and simplebuild_redirect.cfg file exists in search_path directory'
                        f' "{p}" in list entry #{i+1} in item {ctx.item_name} in {ctx.src_descr} (specify'
                        ' the full path to one of the files to specify which one you are interested in using.')
        if p_std is not None:
            res.append(p_std.absolute().resolve())
        else:
            assert p_redirect is not None
            #Ok, load redirection file and expand:
            rd = RedirectionCfg( p_redirect )
            for e in rd.search_path:
                res.append( e )

    return tuple(res)

class DecodeContext:
    def __init__(self, srcdescr, default_dir):
        self.__srcdescr : str = srcdescr
        self.__item_name : str = ''
        self.__default_dir : str = default_dir
    def set_item_name( self, n ):
        self.__item_name = n
    @property
    def src_descr(self) -> str:
        return self.__srcdescr
    @property
    def default_dir(self) -> _OptPath:
        return self.__default_dir
    @property
    def item_name(self) -> str:
        assert self.__item_name
        return self.__item_name

class RedirectionCfg:
    def __init__( self, path : pathlib.Path ):
        try:
            textdata = path.read_text()
        except UnicodeDecodeError:
            error.error(f'Not a text-file: {path}')
        cfgdict = decode_toml_textdata_to_dict( textdata, path )
        if list(cfgdict.keys())!= ['special'] or list(cfgdict['special'].keys())!= ['redirect_search_path']:
            error.error(f'Invalid format of redirection cfg file at {path}')
        ctx = DecodeContext( str(path), path.parent )
        ctx.set_item_name('redirect_search_path')
        self.__sp = decode_list_of_search_paths( ctx,
                                                 cfgdict['special']['redirect_search_path'] )
    @property
    def search_path( self ):
        return self.__sp

class SingleCfg:

    """Class which contains decoded configuration from a single source (a
    simplebuild.cfg file, a python module, etc.)."""

    @classmethod
    def create_from_toml_file( cls, path : pathlib.Path, ignore_build : bool = False ):
        try:
            textdata = path.read_text()
        except UnicodeDecodeError:
            error.error(f'Not a text-file: {path}')
        cfgdict = decode_toml_textdata_to_dict( textdata, path )
        return cls.__create_from_cfgdict( cfgdict, str(path), path.parent, ignore_build = ignore_build )

    @classmethod
    def __create_from_cfgdict( cls, cfgdict : dict, cfg_file : pathlib.Path, default_dir = None, ignore_build : bool = False ):
        o = cls.__create_empty()
        decode_with_schema_and_apply_result_to_obj( cfgdict, o,
                                                    defaultdir = default_dir,
                                                    cfg_file = cfg_file,
                                                    ignore_build = ignore_build )

        if o.project_name != 'core' and 'core' not in o.depend_projects:
            #Always add the core project as a dependency:
            o.depend_projects = list(e for e in o.depend_projects) + ['core']

        o._is_locked = True
        return o

    def __setattr__(self,name,*args,**kwargs):
        if getattr(self,'_is_locked',False):
            error.error(f'Trying to set attribute "{name}" of locked SingleCfg object.')
        object.__setattr__(self,name,*args,**kwargs)

    @classmethod
    def __create_empty( cls ):
        return cls( _private_construct_key = cls.__construct_key )

    __construct_key = id('dummy')#cheap "random" secret integer
    def __init__(self, *, _private_construct_key ):
        self._is_locked = False
        if _private_construct_key != self.__construct_key:
            raise TypeError("Direct initialisation disallowed")

#######################################################################################################

_cache_tomllib = [None]
def import_tomllib():
    if _cache_tomllib[0] is None:
        try:
            import tomllib
        except ModuleNotFoundError:
            try:
                import tomli as tomllib
            except ModuleNotFoundError:
                raise SystemExit('Required toml functionality is not found. Either use '
                                 'python 3.11+ or install tomli'
                                 ' ("conda install -c conda-forge tomli" or "pip install tomli") ')
        _cache_tomllib[0] = tomllib
    return _cache_tomllib[0]

_reexp_valid_identifier = r'^[A-Za-z][A-Za-z0-9_]*$'
_reobj_valid_identifier = re.compile(_reexp_valid_identifier)
def _is_valid_identifier( s ):
    return s and isinstance(s,str) and _reobj_valid_identifier.search(s) is not None

_reexp_valid_lowercase_identifier = r'^[a-z][a-z0-9_]*$'
_reobj_valid_lowercase_identifier = re.compile(_reexp_valid_identifier)
def _is_valid_lowercase_identifier( s ):
    return s and isinstance(s,str) and _reobj_valid_lowercase_identifier.search(s) is not None

is_valid_bundle_name = _is_valid_lowercase_identifier


def _generate_toml_schema():

    def decode_nonempty_str( ctx : TOMLSchemaDecodeContext, item ):
        if not isinstance(item,str) or not str:
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected non-empty string) in {ctx.src_descr}')
        return item

    def decode_valid_identifier_string( ctx : TOMLSchemaDecodeContext, item ):
        if not _is_valid_identifier(item):
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected string matching {_reexp_valid_identifier}) in {ctx.src_descr}')
        return item

    def decode_valid_lowercase_identifier_string( ctx : TOMLSchemaDecodeContext, item ):
        if not _is_valid_lowercase_identifier(item):
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected string matching {_reexp_valid_lowercase_identifier}) in {ctx.src_descr}')
        return item

    def decode_is_bool( ctx : TOMLSchemaDecodeContext, item ):
        if not isinstance( item, bool ):
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected boolean "true" or "false") in {ctx.src_descr}')
        return item

    def decode_is_list_of_valid_identifier_string( ctx : TOMLSchemaDecodeContext, item ):
        if not isinstance( item, list ):
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected list) in {ctx.src_descr}')
        for i,k in enumerate(item):
            if not _is_valid_identifier(k):
                error.error(f'Invalid value of list entry #{i+1} in item {ctx.item_name} (expected string matching {_reexp_valid_identifier}) in {ctx.src_descr}')
        return tuple(item)

    def decode_is_list_of_valid_lowercase_identifier_string( ctx : TOMLSchemaDecodeContext, item ):
        if not isinstance( item, list ):
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected list) in {ctx.src_descr}')
        for i,k in enumerate(item):
            if not _is_valid_lowercase_identifier(k):
                error.error(f'Invalid value of list entry #{i+1} in item {ctx.item_name} (expected string matching {_reexp_valid_lowercase_identifier}) in {ctx.src_descr}')
        return tuple(item)

    def decode_is_env_paths( ctx : TOMLSchemaDecodeContext, item ):
        if not isinstance( item, list ):
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected list) in {ctx.src_descr}')
        res = {}
        for idx,e in enumerate(item):
            parts = list(x.strip() for x in e.split(':<install>/') if x.strip())
            if not len(parts)>=2:
                error.error(f'Invalid value "{e}" for list entry #{idx+1} in item {ctx.item_name} (expected format like "VARNAME:<install>/SUBDIRNAME" but got "{e}") in {ctx.src_descr}')
            varname, install_subdirnames = parts[0],parts[1:]
            if varname not in res:
                res[varname] = set()
            res[varname].update(install_subdirnames)
        return res

    def decode_is_list_of_nonempty_str( ctx : TOMLSchemaDecodeContext, item ):
        if not isinstance( item, list ):
            error.error(f'Invalid value "{item}" for item {ctx.item_name} (expected list) in {ctx.src_descr}')
        for i,k in enumerate(item):
            if not isinstance(k,str) or not k:
                error.error(f'Invalid value of list entry #{i+1} in item {ctx.item_name} (expected non-empty string) in {ctx.src_descr}')
        return tuple(item)

    def decode_str_enum( ctx : TOMLSchemaDecodeContext, item, optionlist ):
        item = decode_nonempty_str( ctx, item )
        if item not in optionlist:
            error.error(f'Invalid option "{item}" for item {ctx.item_name} (must be one of {",".join(optionlist)}) in {ctx.src_descr}')
        return item

    def decode_nonneg_int( ctx : TOMLSchemaDecodeContext, item ):
        if not isinstance(item,int):
            error.error(f'Invalid option "{item}" for item {ctx.item_name} (must be a non-negative integer) in {ctx.src_descr}')
        return item

    def decode_dir( ctx : TOMLSchemaDecodeContext, item ):
        if isinstance( item, pathlib.Path ):
            #special case (likely no longer needed, for sources which already provides Path's rather than str's)
            p = item
            item = str(item)
        else:
            item = decode_nonempty_str( ctx, item )
            p = pathlib.Path(item).expanduser()
        if not p.is_absolute():
            if ctx.default_dir is None:
                error.error(f'Invalid option "{item}" for item {ctx.item_name} (relative paths can only'
                            f' be used in a context where the "root" dir is apparent) in {ctx.src_descr}')
            p = ctx.default_dir / p
        return p.absolute().resolve()

    return dict( project   = dict( name        = (decode_valid_lowercase_identifier_string,None),
                                   pkg_root      = (decode_dir, '.'),
                                   env_paths   = (decode_is_env_paths,[]),
                                  ),
                 depend    = dict( projects     = (decode_is_list_of_valid_lowercase_identifier_string,[]),
                                   search_path = (decode_list_of_search_paths,[])
                                  ),
                 build     = dict( mode = ( lambda a,b : decode_str_enum(a,b,('debug','release')), 'release' ),
                                   #njobs = (decode_nonneg_int, 0),
                                   cachedir = (decode_dir, './simplebuild_cache'),
                                   pkg_filter = (decode_is_list_of_nonempty_str,[]),
                                   #extdep_ignore = (decode_is_list_of_valid_identifier_string,[]),
                                   #cmake_flags = (decode_is_list_of_nonempty_str,[]),
                                   #extra_cflags = (decode_is_list_of_nonempty_str,[]),
                                   #extra_linkflags = (decode_is_list_of_nonempty_str,[]),
                                   #relaxed_compilation = (decode_is_bool,False),
                                  ),
                )

_cache_toml_scheme = [None]
def get_toml_schema():
    if _cache_toml_scheme[0] is None:
        _cache_toml_scheme[0] = _generate_toml_schema()
    return _cache_toml_scheme[0]

def read_text_file( f ):
    try:
        data = f.read_text()
    except UnicodeDecodeError:
        error.error(f'Not a text-file: {f}')
    return data

def decode_toml_textdata_to_dict( textdata : str, path = None ):
    tomllib = import_tomllib()
    descr = path or 'TOML data'
    try:
        cfg = tomllib.loads(textdata)
    except tomllib.TOMLDecodeError as e:
        error.error(f'Syntax error in {descr}: {e}')
    #if not cfg:
    #    error.error(f'No data defined in {descr}. If this is intentional, at least put a single line with the contents "[project]".')
    return cfg

def decode_with_schema_and_apply_result_to_obj( cfg : dict,
                                                targetobj : SingleCfg,
                                                defaultdir,
                                                cfg_file : pathlib.Path,
                                                ignore_build : bool ):
    schema = get_toml_schema()

    ctx = DecodeContext( str(cfg_file), defaultdir )

    #Validate+decode+apply values:
    for section, sectiondata in cfg.items():
        if ignore_build and section=='build':
            continue
        schemadata = schema.get(section)
        #forward compatibility: ignore unknown sections and keys
        if not schemadata:
            error.warn(f'Ignoring unknown simplebuild.cfg section "{section}" of {ctx.src_descr}.')
            continue
        for k,v in sectiondata.items():
            _ = schemadata.get(k)
            if _ is None:
                error.warn(f'Ignoring unknown simplebuild.cfg entry name "{k}" in section "{section}" of {ctx.src_descr}.')
                continue
            decodefct, defval = _
            ctx.set_item_name(f'{section}.{k}')
            v = decodefct(ctx,v)
            setattr( targetobj, f'{section}_{k}', v )
    #Apply default values:
    for sectionname, sectiondata in schema.items():
        if ignore_build and sectionname=='build':
            continue
        for k,(decodefct,defval) in sectiondata.items():
            attrname=f'{sectionname}_{k}'
            if not hasattr(targetobj,attrname):
                ctx.set_item_name(f'DEFAULT::{sectionname}.{k}')
                v = None if defval is None else decodefct(ctx,defval)
                setattr( targetobj,attrname, v )

    setattr(targetobj, '_cfg_file', cfg_file )


    return cfg



"Python module providing access to data files of the various packages"

#This is the python3 version, always returning pathlib paths.

import os
import sys
import pathlib

#We use a little class/sys.modules trick to allow both: Core.FindData(..) and
#Core.FindData.findData(..)

class _themodule:
    def _resolvepp(self,pp):
        return pp.resolve() if pp.exists() else None
    def _resolvepnfn(self,pn,fn):
        _ = pathlib.Path(os.path.join(os.environ['SBLD_DATA_DIR'],pn,fn))
        return self._resolvepp(_)
    def findData(self,pkg,filename=None):
        if filename is not None:
            return self._resolvepnfn(pkg,filename)
        if isinstance(pkg,pathlib.PurePath):
            return self._resolvepp(pkg)#already pathlib path
        #We were given a single argument which is not a pathlib path. Assume it
        #is a string, either in the form of a path, or in the form
        #"<pkgname>/<filename>":
        if not pkg.startswith('/'):
            fs=pkg.split('/')
            if len(fs)==2:
                _=self._resolvepnfn(*fs)
                if _:
                    return _
        return self._resolvepp(pathlib.Path(pkg))
    def __call__(self,pkg,filename=None):
        return self.findData(pkg,filename)
sys.modules[__name__] = _themodule()



def init_project( depbundles = None ):
    #Keywords:
    debug_mode = False
    compact = False
    if 'DEBUG' in depbundles:
        debug_mode = True
        depbundles = [ d for d in depbundles if d!='DEBUG' ]
    if 'COMPACT' in depbundles:
        compact = True
        depbundles = [ d for d in depbundles if d!='COMPACT' ]
    _db = []
    for d in depbundles:
        if d not in _db:
            _db.append(d)
    depbundles = _db

    from .io import print
    import pathlib
    cwd = pathlib.Path.cwd()
    template = ( pathlib.Path(__file__).parent
                 / 'data' / 'cfgtemplate.txt' ).read_text()
    res = ''
    for e in template.splitlines():
        res += e.rstrip()+'\n'
        if debug_mode and e.startswith('[build]'):
            #inject mode statement:
            sbundles = "', '".join(depbundles)
            res += "\n  mode = 'debug'\n\n"
        if depbundles and e.startswith('[depend]'):
            #inject depend.projects list:
            sbundles = "', '".join(depbundles)
            res += f"\n  projects = ['{sbundles}']\n\n"

    if compact:
        res2 = ''
        for e in res.splitlines():
            if '#' in e:
                e = e.split('#',1)[0]
            e = e.rstrip()
            if e:
                res2 += e + '\n'
        res = res2

    for f in cwd.glob('**/simplebuild*.cfg'):
        from .error import error
        error('Can not initialise simplebuild package bundle'
              f' here due to conflicting file: {f}')

    outfile = cwd / 'simplebuild.cfg'
    outfile.write_text(res)
    print(f'Created {outfile.name}. If you wish you can edit it to'
          ' fine-tune your configuration')



# Build script which is intended to make the following steps smooth, painless
# and efficient for users:
#
#  1) Detect environment with cmake
#  2) Configuration step
#  3) Build step
#  4) Install step
#  5) Test launch.
#
# Most importantly, it is intended that minimal cpu time is used to needless
# redo work in steps 1-4 above when user changes code or configuration.

def simplebuild_main( argv = None, prevent_env_setup_msg = False ):

    from . import error
    error.fmt_simplebuild_warnings()#as early as possible
    from . import io as _io
    print = _io.print
    import sys
    import os
    import pathlib

    if argv  is None:
        argv = sys.argv[:]

    progname = os.path.basename( argv[0] )

    from .parse_args import parse_args
    parser, opt = parse_args( argv = argv,
                              return_parser = True )

    if opt.show_version:
        from . import _determine_version
        sys.stdout.write(_determine_version.determine_version()+'\n')
        raise SystemExit

    if opt.quiet:
        _io.make_quiet()
    if opt.verbose:
        _io.make_verbose()

    if opt.env_setup:
        #should actually have been done in _cli.py already
        _io.make_quiet()
        from .envsetup import emit_envsetup
        emit_envsetup()
        raise SystemExit

    if opt.env_unsetup:
        #should actually have been done in _cli.py already
        _io.make_quiet()
        from .envsetup import emit_env_unsetup
        emit_env_unsetup()
        raise SystemExit

    if opt.init is not None:
        from . import init_project
        init_project.init_project( depbundles = opt.init )
        raise SystemExit

    # Now undo the effect of --env-setup on the current environment, so
    # subsequent cmake/make invocations run in a clean environment, not poluted
    # by our own build folder (this also solves
    # https://github.com/mctools/simplebuild/issues/8). But first we note if we
    # need to later on warn the user about their environment needing sb
    # --env-setup:

    if not opt.quiet and not prevent_env_setup_msg:
        from .envsetup import calculate_env_setup
        needs_env_setup = bool(calculate_env_setup())
    else:
        needs_env_setup = False

    from .envsetup import apply_envunsetup_to_dict
    apply_envunsetup_to_dict( os.environ )

    if opt.summary:
        print("FIXME: summary mode is not yet implemented!")
        raise SystemExit

    #setup lockfile:
    from . import dirs

    if not dirs.blddir.exists() and not dirs.installdir.exists():
        #Silently set this in case of completely fresh build, to avoid getting
        #some slightly confusing printouts later:
        opt.insist = True

    if opt.removelock and dirs.lockfile.exists():
        os.remove(dirs.lockfile)

    lockfile_content = str(os.getpid())
    if dirs.lockfile.exists():
        locking_pid = dirs.lockfile.read_text()
        error.error('Presence of lock file indicates competing invocation of '
                    f'{progname} (by pid {locking_pid}). Force removal'
                    f' with {progname} --removelock if you are sure this is incorrect.')
    dirs.create_bld_dir()
    dirs.lockfile.write_text(lockfile_content)

    assert dirs.lockfile.read_text()==lockfile_content

    def unlock():
        expected_content = lockfile_content
        if ( dirs.lockfile.exists()
             and dirs.lockfile.read_text()==expected_content ):
            dirs.lockfile.unlink()
    import atexit
    atexit.register(unlock)

    from . import conf
    from . import envcfg

    if opt.clean:
        if dirs.blddir.is_dir() or dirs.installdir.is_dir():
            if not opt.quiet:
                print("Removing temporary cache directories.")
            conf.safe_remove_install_and_build_dir()
            #Let us perhaps also try and remove the cachedir, i.e. the parent
            #dir of install and build. We want to remove it to not litter an
            #empty "simplebuild_cache" in peoples bundle dirs, but we also do
            #not want to remove a directory which a user might have created
            #manually for their cache (even if empty, since it might still be
            #confusing). As a compromise, we remove it only if it is in the
            #default location with the default name (users messing with pkg_root
            #or cachedir are anyway expected to be more advanced):
            dcache = envcfg.var.build_dir_resolved.parent
            assert dcache == envcfg.var.install_dir_resolved.parent
            if ( dcache == ( envcfg.var.projects_dir / 'simplebuild_cache' )
                 and not any( dcache.iterdir() ) ):
                 dcache.rmdir()
        else:
            if not opt.quiet:
                print("Nothing to clean. Exiting.")
        raise SystemExit

    #Detect changes to system cmake or python files and set opt.examine or opt.insist as appropriate.
    from . import mtime
    from . import utils
    systs = (mtime.mtime_cmake(),mtime.mtime_pymods())
    try:
        oldsysts = utils.pkl_load(dirs.systimestamp_cache)
    except IOError:
        if not opt.insist:
            if dirs.systimestamp_cache.exists():
                print("Could not load timestamp cache insist due to IOError loading dirs.systimestamp_cache")
            opt.insist=True
        oldsysts = (None,None)

    if envcfg.var.allow_sys_dev:
        #prevent changes in system from triggering rebuilds (for careful system
        #development use):
        systs = oldsysts

    if not opt.insist and oldsysts!=systs:
        if oldsysts[0]!=systs[0]:
            opt.examine = True
        if not opt.insist and oldsysts[1]!=systs[1]:
            print("simple-build-system time stamp changed -> performing complete rebuild for safety.")
            opt.insist = True

    if not opt.insist and dirs.envcache.exists():
        envdict=utils.pkl_load(dirs.envcache)
        #insist rebuilding COMPLETELY from scratch if install or build dirs changed:
        _autoreconf_inst_dir = envdict['_autoreconf_environment'].get('install_dir')
        _autoreconf_bld_dir = envdict['_autoreconf_environment'].get('build_dir')
        if _autoreconf_inst_dir != str(conf.install_dir()):
            print("Performing complete rebuild since the cache install dir changed (%s -> %s)"%(_autoreconf_inst_dir,
                                                                                                str(conf.install_dir())))
            opt.insist = True
        elif _autoreconf_bld_dir != str(conf.build_dir()):
            print("Performing complete rebuild since the cache build dir changed (%s -> %s)"%(_autoreconf_bld_dir,
                                                                                              str(conf.build_dir())))
            opt.insist = True

    if opt.insist:
        conf.safe_remove_install_and_build_dir()
        dirs.create_bld_dir()

    utils.pkl_dump(systs,dirs.systimestamp_cache)

    select_filter = envcfg.var.pkg_filter
    assert select_filter is not None

    from . import backend

    err_txt,unclean_exception = None,None
    error.default_error_type = error.Error
    try:
        pkgloader = backend.perform_configuration(select_filter=select_filter,
                                                  force_reconf=opt.examine,
                                                  load_all_pkgs = opt.query_mode,
                                                  quiet=opt.quiet,
                                                  verbose=opt.verbose)
    except KeyboardInterrupt:
        err_txt = "Halted by user interrupt (CTRL-C)"
    except error.CleanExit as ce:
        #errors already handled, exit directly:
        sys.exit(ce._the_ec)
    except Exception as e:
        err_txt=str(e)
        if not err_txt:
            err_txt='<unknown error>'
        if not isinstance(e,error.Error):
            unclean_exception = e
    except SystemExit as e:
        if str(e)!="knownhandledexception":
            err_txt = "Halted by unexpected call to system exit!"
            unclean_exception = e

    error.default_error_type = SystemExit

    if err_txt:
        #fixme: unprefixed printouts?:
        print("\n\nERROR during configuration:\n\n  %s\n\n"
              "Aborting."%(err_txt.replace('\n','\n  ')))
        #make all packages need reconfig upon next run:
        from . import db
        db.db['pkg2timestamp']={}
        db.save_to_file()
        #exit (with ugly traceback if unknown type of exception):
        if unclean_exception:
            error.print_traceback(unclean_exception)
        sys.exit(1)

    assert dirs.makefiledir.is_dir()

    def query_pkgs():
        #returns list of (pkg,filenames) where filenames is None, or a list of
        #the files in the pkg to search (e.g. ['pkg.info','pycpp_bla/mod.cc']
        all_pkgs = list(sorted(pkgloader.pkgs))
        if not opt.querypaths:
            return [(p,None) for p in all_pkgs]
        res=[]
        for p in all_pkgs:
            search_entire_package = False
            filenames=[]
            for qp in opt.querypaths:
                assert isinstance(p.dirname,str)#fixme: to pathlib.Path
                pdir = pathlib.Path(p.dirname)
                if pdir == qp or utils.path_is_relative_to( pdir, qp ):
                    search_entire_package = True
                    break
                if utils.path_is_relative_to( qp, pdir ):
                    filenames.append( qp.relative_to(pdir) )
            if search_entire_package:
                res.append( (p, None) )
            elif filenames:
                res.append( (p, filenames) )
        return res

    if opt.grep:
        qp = query_pkgs()
        print("Grepping %i packages for pattern \"%s\""%(len(qp),opt.grep))
        print()
        n=0
        from . import grep
        for pkg,filenames in qp:
            n += grep.grep( pkg,
                            opt.grep,
                            filenames = filenames,
                            countonly = opt.grepc )
        print()
        print("Found %i matches"%n)
        raise SystemExit

    if opt.replace:
        qp = query_pkgs()
        pattern = opt.replace
        from . import replace
        search_pat, replace_pat = replace.decode_pattern(pattern)
        if not search_pat:
            parser.error("Bad syntax in replacement pattern: %s"%pattern)
        print()
        print("Replacing all \"%s\" with \"%s\""%(search_pat,replace_pat))
        n = 0
        for pkg,filenames in qp:
            n += replace.replace( pkg,
                                  search_pat,
                                  replace_pat,
                                  filenames = filenames )
        print()
        print("Performed %i replacements"%n)
        raise SystemExit

    if opt.find:
        qp = query_pkgs()
        pattern = opt.find
        from . import find
        print()
        print("Finding files and paths matching \"%s\""%(opt.find))
        print()
        n = 0
        for pkg,filenames in qp:
            n += find.find( pkg,
                            pattern = opt.find,
                            filenames = filenames )
        print()
        print("Found %i matches"%n)
        raise SystemExit

    if opt.incinfo:
        import glob
        def _val_incinfofn(fn):
            if '*' in fn:
                #try to expand wildcards:
                fff = sorted(glob.glob(fn))
                if fff:
                    return sum((_val_incinfofn(ff) for ff in fff),[])
            if not os.path.exists(fn):
                #Could be of form "pkgname/subdir/file. If so, expand pkgname part
                #to full path to package:
                i=fn.split(os.path.sep)
                pkg=pkgloader.name2pkg.get(i[0],None) if len(i)==3 else None
                if pkg:
                    return _val_incinfofn(dirs.pkg_dir(pkg,i[1],i[2]))
                else:
                    parser.error("File not found: %s"%fn)
            if os.path.isdir(fn):
                parser.error("Not a file: %s"%fn)
            fn=os.path.abspath(os.path.realpath(fn))
            p = pathlib.Path(fn).absolute().resolve()
            simplebuild_pkg_dirs = [dirs.projdir, *dirs.extrapkgpath]
            if not any( utils.path_is_relative_to(p,d)
                        for d in simplebuild_pkg_dirs):
                _dirsfmt=('\n '.join(str(e) for e in simplebuild_pkg_dirs))
                parser.error(f'File {p} must be located under one of the '
                             f'following directories:\n{_dirsfmt}')
            return [fn]#expands to a single file
        from . import incinfo
        fnsraw = opt.incinfo.split(',') if ',' in opt.incinfo else [opt.incinfo]
        fns = sum((_val_incinfofn(fnraw) for fnraw in fnsraw),[])
        #remove duplicates (relies on seen.add returning None)
        seen=set()
        fns = [fn for fn in fns if not (fn in seen or seen.add(fn))]
        #Dispatch to backend:
        if len(fns)==1:
            incinfo.provide_info(pkgloader,fns[0])
        else:
            incinfo.provide_info_multifiles(pkgloader,fns)
        raise SystemExit

    if opt.pkginfo:
        pkg=pkgloader.name2pkg.get(opt.pkginfo,None)
        if not pkg:
            utils.err('Unknown package "%s"'%opt.pkginfo)
        else:
            pkg.dumpinfo(pkgloader.autodeps)
            raise SystemExit

    if opt.pkggraph:
        dotfile=dirs.blddir / 'pkggraph.dot'
        from . import dotgen
        dotgen.dotgen(pkgloader,dotfile,enabled_only=opt.pkggraph_activeonly)
        if not opt.quiet:
            print('Package dependencies in graphviz DOT format has'
                  ' been generated in %s'%(dotfile))
        ec=utils.system('dot -V > /dev/null 2>&1')
        if ec:
            if not opt.quiet:
                print('Warning: command "dot" not found or ran into problems.')
                print('Please install graphviz to enable graphical'
                      ' dependency displays')
            sys.exit(1)
        pkggraphout = dirs.blddir / 'pkggraph.png'
        import shlex
        ec=utils.system(
            'unflatten -l3 -c7 %s|dot -Tpng -o %s'%(shlex.quote(str(dotfile)),
                                                    shlex.quote(str(pkggraphout)))
        )
        if ec or not pkggraphout.is_file():
            if not opt.quiet:
                print('Error: Problems with dot command while transforming'
                      f' {dotfile.name} to {pkggraphout.name}')
            sys.exit(1)
        else:
            if not opt.quiet:
                print('Package dependencies in PNG format has been generated'
                      ' in %s'%pkggraphout)
        raise SystemExit


    if not opt.njobs:
        from . import cpudetect
        opt.njobs=cpudetect.auto_njobs()

    if not opt.quiet:
        print("Configuration completed => Launching build with %i parallel processes"%opt.njobs)

    #VERBOSE:
    # -1: always quiet
    #  0: only warnings
    #  1: extra verbose printouts
    if opt.verbose:
        extramakeopts=' VERBOSE=1'
    elif opt.quiet:
        extramakeopts=' VERBOSE=-1'
    else:
        extramakeopts=''

    assert dirs.makefiledir.is_dir()

    ec = utils.system(f'cd {dirs.makefiledir} && '
                      'make --warn-undefined-variables -f Makefile'
                      f' -j{opt.njobs}{extramakeopts}')
    if ec!=0:
        if not opt.quiet:
            print("ERROR: Build problems encountered")
        sys.exit(1 if ec > 128 else ec)

    if not opt.quiet:
        from . import build_summary
        build_summary.produce_build_summary(
            pkgloader = pkgloader,
            verbose = opt.verbose
        )

    if opt.runtests:
        assert (conf.test_dir().parent / '.sbbuilddir').exists()
        import shutil
        shutil.rmtree(conf.test_dir(),ignore_errors=True)
        _testfilter=''
        if opt.testfilter:
            _testfilter=[fltr.strip() for fltr in opt.testfilter.split(',') if fltr.strip()]
        from .testlauncher import perform_tests
        ec = perform_tests( testdir = dirs.testdir,
                            installdir = dirs.installdir,
                            njobs = opt.njobs,
                            nexcerpts = opt.nexcerpts,
                            filters = _testfilter,
                            do_pycoverage = False,
                            pkgloader = pkgloader )

        from . import env
        from . import col
        cp = env.env['cmake_printinfo']
        if ec==0 and (cp['unused_vars'] or cp['other_warnings']):
            #Make sure user sees these warnings:
            print('%sWARNING%s There were warnings (see above)'%(col.bad,col.end))
            print()
        if ec:
            sys.exit(ec)

    if not opt.quiet:
        if needs_env_setup:
            from . import col
            print(f'{col.warnenvsetup}Build done. To use the resulting environment you must first enable it!{col.end}')
            print()
            print(f'{col.warnenvsetup}Type the following command (exactly) to do so (undo later by --env-unsetup instead):{col.end}')
            print()
            print(f'    {col.warnenvsetup}eval "$({progname} --env-setup)"{col.end}')
            print()
        else:
            print("Build done. You are all set to begin using the software!")
            print()
            if not needs_env_setup:
                print(f'To see available applications, type "{conf.runnable_prefix}" and hit the TAB key twice.')
                print()




def setup(app):
    """Install the plugin."""
    from ._determine_version import determine_version
    version = determine_version()

    app.add_role('sbpkg', sbpkg_role)
    app.add_config_value('simplebuild_pkgbundles', None, 'env')
    return { 'version' : version }

#FIXME: optional url-exists checking? For the CI mainly I guess.

def sbpkg_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """Link to the source code of a given simplebuild package."""
    from .pkgutils import pkgname_valid_or_errmsg
    app = inliner.document.settings.env.app
    def _err(msg = None):
        _msg = inliner.reporter.error( msg or 'Syntax error in: "%s"'%text,
                                       line = lineno)
        return ( [inliner.problematic(rawtext, rawtext, _msg)],
                 [_msg] )

    linkinput = text.strip()
    override_text = None
    if '<' in linkinput:
        _parts = [e.strip() for e in linkinput.split('<')]
        if not len(_parts)==2 or not _parts[1].endswith('>'):
            return _err()
        _parts[1] = _parts[1][:-1].strip()
        _parts = [e.strip() for e in _parts if e.strip()]
        if len(_parts)!=2 or not all( (e and '>' not in e) for e in _parts ):
            return _err()
        override_text, linkinput = _parts

    _parts = [p.strip() for p in linkinput.split('/',1) if p.strip()]
    if len(_parts) not in (1,2):
        return _err()
    pkgname,subpath = (_parts[0],None) if len(_parts) == 1 else _parts
    _ = pkgname_valid_or_errmsg( pkgname )
    if _ is not None:
        return _err(_)
    sbpkgs = load_sbpkgs(app)
    if sbpkgs['error_msg'] is not None:
        return _err(sbpkgs['error_msg'])
    pkginfo = sbpkgs['pkgs'].get(pkgname)
    if not pkginfo:
        return _err('Could not find package named "%s"'%pkgname)
    url = pkginfo['pkgdir_url']
    linktext = pkgname

    #Note: github links should use /blob/ for files and /tree/ for dirs, so
    #allow input cfg to specify '[blob|tree]' in the relevant location by
    #performing the relevant .replace's below:

    if subpath:
        _=pkginfo['pkgdir_local'] / subpath
        _is_file = _.is_file()
        _is_dir = _.is_dir()
        if not (_is_file or _is_dir):
            return _err(f'File not found in package: {_}')
        if _is_file:
            _gh_blobortree = 'blob'
        url = url.replace('[blob|tree]',
                          'blob' if _is_file else 'tree')
        url += '/%s'%subpath
        linktext += '/%s'%subpath
    else:
        url = url.replace('[blob|tree]','tree')

    if override_text:
        linktext = override_text

    node = _create_link_node(rawtext, app, linktext, url, options)
    return [node], []

def _create_link_node(rawtext, app, linktext, linkurl, options):
    from docutils import nodes as _du_nodes
    from docutils.parsers.rst.roles import set_classes as _du_set_classes
    _du_set_classes(options)
    return _du_nodes.reference(rawtext, linktext, refuri=linkurl,
                               **options)

_cache = [False,None]
def load_sbpkgs( app ):
    if not _cache[0]:
        _cache[1] = _actual_load_sbpkgs( app )
        _cache[0] = True
    return _cache[1]

def _actual_load_sbpkgs( app ):
    try:
        bundles = app.config.simplebuild_pkgbundles
        if not bundles:
            raise AttributeError
    except AttributeError as e:
        raise ValueError('simplebuild_pkgbundles configuration value'
                         ' is not set (%s)' % str(e))

    from .pkgutils import find_pkg_dirs_under_basedir as _findpkgs
    errors = []
    pkgs = {}
    def _errfct( msg ):
        errors.append( msg )
    for n,( pkgroot, urlbase ) in bundles.items():
        assert ( pkgroot / 'simplebuild.cfg' ).is_file()
        for p in _findpkgs( pkgroot, error_fct = _errfct ):
            if errors:
                break
            assert ( p / 'pkg.info' ).is_file()
            if p.name in pkgs:
                _errfct('Duplicate package name: "%s"'%p.name)
                break
            url = urlbase + '/' + str( p.relative_to(pkgroot) )
            pkgs[ p.name ] = { 'name' : p.name,
                               'pkgdir_local' : p,
                               'pkgdir_url' : url }
        if errors:
            break
    return dict( error_msg = errors[0] if errors else None,
                 pkgs = None if errors else pkgs )





# We have per-package make-files and targets. We also have a global DB to
# communicate header files and libraries between packages. In the same file as the
# global DB is in, we also keep info about the "last change" timestamps of
# packages as well as a list of previously enabled packages.
#
# All target objects must support pickling/depickling after their __init__ calls.
#
# * When neither a package nor any package it depends on has changed files, nothing
#   package specific is done and the makefile is left as it was.
#
# * When a package has changed files, its targets are completely recreated (and
#   pickled for future use). Their setup() calls are then done.
#
# * When a package has no changed files but there are changes in a pkg it depends
#   on, its targets are depickled and only setup() calls and the Makefile are redone.
#
# Global targets are also added, for now this is just a target which makes sure
# that the symlinks in ${INST}/include are up to date (NB: different symlinks are
# used for the actual build!)
#
# Thus, a target should implement two methods:
# #   __init__: Sets up object instance and perform as much work as can be done without knowing details of other packages.
# #             Also puts info needed by other pkgs into global db (libinc files + presence of libsrc)
# #             (remember to prune the global db when package is reconfed)
# #     <---- pickling happens here ---->
# #   setup: Performs work which depends on files in other packages.
#
#
#Derived classes should only keep members which can be pickled and will be valid
#in a subsequent invocation (i.e. don't store pkg objects!).
#
#It is OK to keep other target instances, but only if they are from the same package.

from . import col

class Target:
    contains_message=False#todo: we should probably have a custom message in all targets
    pkglevel = True#set to false if other targets in the pkg depends on this
    isglobal = False

    def setup(self,pkg):
        #pkg==None if global target
        pass

    def validate(self):
        assert self.name#unique name (file or just a label)
        assert isinstance(self.name,str)
        if self.isglobal:
            assert self.pkgname is None
        else:
            assert self.pkgname
            assert isinstance(self.pkgname,str)
        assert isinstance(self.deps,list)#names of other targets which we depend on
        assert isinstance(self.code,list)#The code to run

class TargetPkgDone(Target):
    def __init__(self,pkg,pkgtargets):
        self.name=pkg.name
        self.pkgname=pkg.name
        self.deps=[t.name for t in pkgtargets if t.pkglevel]
        self.contains_message = True
        self.code = ['@if [ ${VERBOSE} -ge 0 ]; then echo "%sPackage %s done%s"; fi'%(col.bldcol('pkg'),self.pkgname,col.bldend)]



import os
import pathlib
from . import conf
from . import target_base
from . import env
from . import dirs
from . import langs
from . import utils
from . import includes
from . import col
from . import db
from . import error
from . import envcfg
from . import tfact_headerdeps

join=os.path.join

has_realpath = os.path.exists('/usr/bin/realpath')

class TargetBinaryObject(target_base.Target):
    def __init__(self,pkg,subdir,bn,e,lang,shlib,cflags,possible_privincs):
        is_py_mod = subdir.startswith('pycpp_')#todo better way?
        langinfo=env.env['system']['langs'][lang]
        self.pkglevel=False#needed by the corresponding TargetBinary object
        self.pkgname=pkg.name
        self._fn=bn+e
        self._lang=lang
        self._shlib=shlib
        d=dirs.makefile_pkg_cache_dir(pkg,'objs',subdir)
        self.name=join(d,langinfo['name_obj']%bn)
        fn=bn+e
        sf=dirs.makefile_pkg_dir(pkg,subdir,fn)
        self.__pklfile=dirs.pkg_cache_dir(pkg,'objs',subdir,fn+'_deps.pkl')
        self.deps=[dirs.makefile_blddir(self.__pklfile)]
        self.deps+=['${LANG}/%s'%lang,'%s_prepinc'%pkg.name,sf]#FIXME perhaps on /cxxpch in case of cxx? (if has pch reqs)
        extra_flags=['-I%s'%dirs.makefile_pkg_cache_dir(pkg,'inc'),'-DPACKAGE_NAME=%s'%pkg.name]

        if cflags:
            extra_flags+=cflags
        if shlib:
            extra_flags+=['-fPIC']#Could we get this from cmake already??
        for extdep in pkg.extdeps():
            extra_flags+=['${CFLAGS_%s_%s}'%(('CXX' if lang=='cxx' else 'C'),extdep)]
            self.deps+=['${EXT}/%s'%extdep]

        if is_py_mod:
            assert shlib
            extra_flags.append( '${PYBIND11_MODULE_CFLAGS}' )
        else:
            extra_flags.append( '${PYBIND11_EMBED_CFLAGS}' )

        if pkg.extraflags_comp:
            extra_flags+=pkg.extraflags_comp

        #contains_message=True
        obj_create_key = 'create_obj_for_shlib' if shlib else 'create_obj_for_exe'

        self.code=['@if [ ${VERBOSE} -ge 0 ]; then echo "  %sBuilding %s/%s/%s%s"; fi'%(col.bldcol('objectfile'),
                                                                                        pkg.name,
                                                                                        subdir,
                                                                                        os.path.basename(self.name),
                                                                                        col.bldend),
                   'mkdir -p %s'%d,
                   langinfo[obj_create_key]%(' '.join(extra_flags),'`realpath %s`'%sf if has_realpath else sf,self.name)]

        priv,pub = includes.find_includes(dirs.pkg_dir(pkg,subdir,fn),pkg)

        # NOTE THAT FOLLOWING LINES OF CODE ARE SIMILAR TO ONES IN tfact_headerdeps.py !!!

        #deal with the private dependencies as well as public deps to same pkg
        #here, postpone the other public ones until setup() when other packages are available.
        self._pkloffset=len(self.deps)
        if priv and possible_privincs:
            for p in priv:
                if p in possible_privincs:
                    self.deps+=['${TRG}/%s__%s__%s'%(pkg.name,subdir,p)]
        if pub:
            pkg.update_incs()
            incs=db.db['pkg2inc'][pkg.name]
            self.deps += ['${TRG}/%s__libinc__%s'%(p,f) for (p,f) in pub if (p==self.pkgname and f in incs)]
            self._pub = [(p,f) for (p,f) in pub if p!=self.pkgname]
        else:
            self._pub=None

    def setup(self,pkg):
        if self._pub:
            # NOTE THAT FOLLOWING LINES OF CODE ARE SIMILAR TO ONES IN tfact_headerdeps.py !!!
            p2i=db.db['pkg2inc']
            d=pkg.deps_names()
            self.deps+=['${TRG}/%s__libinc__%s'%(p,f) for (p,f) in self._pub if (p in d and f in p2i.get(p,[]))]
        pklcont=set(self.deps[self._pkloffset:])
        utils.mkdir_p(os.path.dirname(self.__pklfile))
        utils.update_pkl_if_changed(pklcont,self.__pklfile)

class TargetBinary(target_base.Target):

    def __init__(self,pkg,subdir,lang,shlib,object_targets,instsubdir,namefct,descrfct=None,checkfct=None):
        is_py_mod = subdir.startswith('pycpp_')#todo better way?
        db.db['pkg2parts'][pkg.name].add(subdir)
        langinfo=env.env['system']['langs'][lang]
        self.pkgname=pkg.name
        self._subdir=subdir

        pattern='name_lib' if shlib else 'name_exe'
        filename=namefct(pkg,subdir,langinfo[pattern])
        if checkfct:
            checkerrmsg = checkfct(pkg,subdir,filename)
            if checkerrmsg:
                error.error(checkerrmsg)

        if subdir.startswith('app_'):
            assert not shlib
            pkg.register_runnable(filename)
            db.db['pkg2runnables'].setdefault(pkg.name,set()).add(filename)

        dcol,descr = descrfct(pkg,subdir,filename) if descrfct else ('','%s/%s'%(pkg.name,filename))
        dcolend = col.bldend if dcol else ''
        self.name='%s_%s'%(pkg.name,subdir)

        self.deps=[]
        for ot in object_targets:
            self.deps.append(ot.name)
        self.deps.sort()
        sobjs=' '.join(self.deps)
        cachedir=dirs.pkg_cache_dir(pkg,'objs')
        utils.mkdir_p(cachedir)
        objlistfile=join(cachedir,'objs_%s.txt'%subdir)#to make sure we rebuild when removing a src file
        self.deps+=[dirs.makefile_blddir(objlistfile)]#'%s_prepinc'%pkg.name]
        extra_flags=[]
        extra_flags+=['%s']#This is used in self.setup() to add in library deps.
        for extdep in pkg.extdeps():
            extra_flags+=['${LDFLAGS_%s} ${LDFLAGS_%s_%s_%s}'%(extdep,extdep,('LIB' if shlib else 'EXE'),lang)]
        if pkg.extraflags_link:
            extra_flags+=pkg.extraflags_link
        rpath_pattern = langinfo['rpath_flag_lib' if shlib else 'rpath_flag_exe']
        def append_to_rpath( extra_flags, p ):
            extra_flags += [ rpath_pattern%str(p) ]
            if langinfo['can_use_rpathlink_flag'] and '-rpath' in rpath_pattern and '-rpath-link' not in rpath_pattern:
                extra_flags += [ rpath_pattern.replace('-rpath','-rpath-link')%str(p) ]
        append_to_rpath( extra_flags, join('${INST}','lib') )
        append_to_rpath( extra_flags, join('${INST}','lib','links') )

        if is_py_mod:
            assert shlib
            extra_flags.append( '${PYBIND11_MODULE_LDFLAGS}' )
        else:
            extra_flags.append( '${PYBIND11_EMBED_LDFLAGS}' )

        conda_prefix =  envcfg.var.conda_prefix
        if conda_prefix:
            _cp = pathlib.Path(conda_prefix) / 'lib'
            if _cp.is_dir():
                append_to_rpath( extra_flags, _cp)
        if not os.path.exists(objlistfile) or open(objlistfile).read()!=sobjs:
            open(objlistfile,'w').write(sobjs)
        d=join('${INST}',instsubdir)
        #contains_message=True
        pattern = 'create_lib' if shlib else 'create_exe'
        if shlib and '-fPIE' in extra_flags:
            #-fPIE is for position independent executables, not libraries:
            extra_flags = [e for e in extra_flags if e!='-fPIE']
        self.code=['@if [ ${VERBOSE} -ge 0 ]; then echo "  %sCreating %s%s"; fi'%(dcol,descr,dcolend),
                   'mkdir -p %s'%d,
                   langinfo[pattern]%(' '.join(extra_flags),
                                      sobjs,
                                      join(d,filename))]

    def setup(self,pkg):
        l=[]
        if self._subdir!='libsrc':
            if pkg.has_lib():
                l+=[conf.libldflag(pkg)]
            self.deps+=['%s_libavail'%self.pkgname]
        for p in pkg.deps():
            if p.has_lib():
                l+=[conf.libldflag(p)]
            self.deps+=['%s_libavail'%p.name]
        if l:
            l.insert(0, '-L${INST}/lib')
        self.code[-1]=self.code[-1]%(' '.join(l))

def create_tfactory_binary(instsubdir=None,pkglib=False,shlib=False,allowed_langs=None,namefct=None,flagfct=None,descrfct=None,checkfct=None):
    if not allowed_langs:
        allowed_langs = langs.langs
    if pkglib:
        shlib=True
        instsubdir='lib'
    blainstsubdir=instsubdir#weird that we need this to propagate the variables into the next function

    def tfact_bin(pkg,subdir):
        instsubdir=blainstsubdir
        if '%s' in instsubdir:
            instsubdir = instsubdir%pkg.name
        srcs=[]
        hdrs=[]
        langspresent=set()
        skip_fct = None

        for f in utils.listfiles(dirs.pkg_dir(pkg,subdir),ignore_logs=subdir.startswith('app_')):
            n,e=os.path.splitext(f)
            assert n
            langsrc = langs.srcext2lang.get(e,None)
            if langsrc:
                langspresent.add(langsrc)
                if skip_fct and skip_fct(n):
                    continue
                srcs+=[(n,e,langsrc)]
                continue
            langhdr = None if langsrc else langs.hdrext2lang.get(e,None)
            if langhdr:
                langspresent.add(langhdr)
                hdrs+=[(n,e,langhdr)]
                continue
            error.error("File is neither src or header file of allowed language: %s"%f)

        if not srcs:
            error.error("No source files found in %s/%s (remove this directory if not needed)"%(pkg.name,subdir))
        if len(langspresent)!=1:
            assert len(langspresent)>1
            error.error("Files for multiple languages found in %s/%s"%(pkg.name,subdir))
        langspresent=langspresent.pop()
        if langspresent not in allowed_langs:
            error.error("Files of language '%s' found in %s/%s where only allowed languages are '%s'"%(langspresent,
                                                                                                       pkg.name,subdir,
                                                                                                       ';'.join(allowed_langs)))
        tgts=[]
        extra_cflags=flagfct(pkg,subdir) if flagfct else []

        possible_privincs = langs.headers_in_dir(dirs.pkg_dir(pkg,subdir))
        for bn,e,lang in srcs:
            tgts+=[TargetBinaryObject(pkg,subdir,bn,e,lang,
                                      shlib=shlib,cflags=extra_cflags,
                                      possible_privincs=possible_privincs)]
        tb=TargetBinary(pkg,subdir,langspresent,
                        shlib=shlib,object_targets=tgts,
                        instsubdir=instsubdir,namefct=namefct,descrfct=descrfct,
                        checkfct=checkfct)
        if pkglib:
            pkg.libtarget=tb
        tgts+=[tb]

        tgts += tfact_headerdeps.tfactory_headerdeps(pkg,subdir)

        return tgts
    return tfact_bin





import os
from . import target_base
from . import utils
from . import dirs
from . import conf
from . import col
from . import error
from . import db
join = os.path.join
basename=os.path.basename

pypkgname = __name__.split('.')[0]
assert pypkgname != '__main__'

class TargetRefLogs(target_base.Target):
    def __init__(self,pkg,applogs,scriptlogs):
        #No need for this, since each package will always have this part:
        #db.db['pkg2parts'][pkg.name].add('testref_links')
        targetdir=dirs.installdir / 'tests/testref'
        utils.mkdir_p(targetdir)
        neededlinks=set()
        for f in applogs:
            tn = '%s.log'%conf.runnable_name(pkg,basename(os.path.dirname(f))[4:])
            neededlinks.add((targetdir,tn,f))
        for f in scriptlogs:
            tn = '%s.log'%conf.runnable_name(pkg,basename(f)[:-4])
            neededlinks.add((targetdir,tn,f))
        d=dirs.pkg_cache_dir(pkg,'testref')
        utils.mkdir_p(d)
        pklfile=join(d,'testref.pkl')
        mpf=dirs.makefile_blddir(pklfile)
        utils.update_pkl_if_changed(neededlinks,pklfile)
        for _,tn,_ in neededlinks:
            db.db['pkg2reflogs'].setdefault(pkg.name,set()).add(tn)

        self.name='%s_reflogs'%pkg.name
        self.pkgname=pkg.name
        self.deps=[mpf]
        self.code = ['@if [ ${VERBOSE} -ge 0 ]; then echo "  %sUpdating symlinks %s/testlogs%s"; fi'%(col.bldcol('symlink'),pkg.name,col.bldend),
                     'python3 -m%s.instsl2 %s ${VERBOSE}'%(pypkgname,mpf)]

def tfactory_reflogs(pkg,dirtypes):
    #Must complain about any .log file in app_* which is not test.log
    #Must complain about any .log file in scripts/ which does not have an associated script

    applogs,scriptlogs=[],[]
    if 'app_' in dirtypes:
        applogs = [f for f in dirs.pkg_dir(pkg).glob('app_*/*.log') if not conf.ignore_file(f)]
        for f in applogs:
            if basename(f)!='test.log' or utils.is_executable(f):
                if basename(f)!='test.log':
                    e='Log files in app_* dirs can only be named "test.log".'
                if utils.is_executable(f):
                    e='Log files should not be executable (correct by running chmod -x <filename>).'
                error.error('%s\n\nProblematic file: %s'%(e,f))
    if 'scri' in dirtypes:
        scriptlogs = [f for f in dirs.pkg_dir(pkg,'scripts').glob('*.log') if not conf.ignore_file(f)]
        for f in scriptlogs:
            sf,ext = os.path.splitext(f)
            assert ext=='.log'
            if not os.path.exists(sf):
                error.error('Log file in script dir does not have a corresponding script!\n',
                            'Problematic file: %s'%f,
                            'Missing script: %s'%sf)
            if utils.is_executable(f):
                error.error('Log files should not be executable (correct by running chmod -x <filename>)!\n',
                            'Problematic file: %s'%f)
    #always create since we want to also undo links:
    #if not applogs and not scriptlogs:
    #    return []
    return [TargetRefLogs(pkg,applogs,scriptlogs)]
